# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"plot_sequence": "10_core_overview.ipynb",
         "plot_sequence_1d": "10_core_overview.ipynb",
         "get_alphabet": "11_core_elements.ipynb",
         "get_element_counts": "11_core_elements.ipynb",
         "get_first_positions": "11_core_elements.ipynb",
         "get_element_frequency": "11_core_elements.ipynb",
         "plot_element_counts": "11_core_elements.ipynb",
         "get_subsequences": "12_core_subsequences.ipynb",
         "get_ndistinct_subsequences": "12_core_subsequences.ipynb",
         "get_unique_ngrams": "13_core_ngrams.ipynb",
         "get_all_ngrams": "13_core_ngrams.ipynb",
         "get_ngram_universe": "13_core_ngrams.ipynb",
         "get_ngram_counts": "13_core_ngrams.ipynb",
         "plot_ngram_counts": "13_core_ngrams.ipynb",
         "get_transitions": "14_core_transitions.ipynb",
         "get_ntransitions": "14_core_transitions.ipynb",
         "get_transition_matrix": "14_core_transitions.ipynb",
         "plot_transition_matrix": "14_core_transitions.ipynb",
         "get_spells": "15_core_spells.ipynb",
         "get_longest_spell": "15_core_spells.ipynb",
         "get_spell_durations": "15_core_spells.ipynb",
         "is_recurrent": "16_core_statistics.ipynb",
         "get_entropy": "16_core_statistics.ipynb",
         "get_turbulence": "16_core_statistics.ipynb",
         "get_complexity": "16_core_statistics.ipynb",
         "get_routine": "16_core_statistics.ipynb",
         "plot_sequences": "20_multi_overview.ipynb",
         "are_recurrent": "21_multi_attributes.ipynb",
         "get_summary_statistic": "21_multi_attributes.ipynb",
         "get_routine_scores": "21_multi_attributes.ipynb",
         "get_synchrony": "21_multi_attributes.ipynb",
         "get_sequence_frequencies": "21_multi_attributes.ipynb",
         "get_motif": "22_multi_derivatives.ipynb",
         "get_modal_state": "22_multi_derivatives.ipynb",
         "get_optimal_distance": "23_multi_edit_distances.ipynb",
         "get_levenshtein_distance": "23_multi_edit_distances.ipynb",
         "get_hamming_distance": "23_multi_edit_distances.ipynb",
         "get_combinatorial_distance": "24_multi_nonalignment.ipynb"}

modules = ["core/__init__.py",
           "core/elements.py",
           "core/subsequences.py",
           "core/ngrams.py",
           "core/transitions.py",
           "core/spells.py",
           "core/statistics.py",
           "multi/__init__.py",
           "multi/attributes.py",
           "multi/derivatives.py",
           "multi/editdistances.py",
           "multi/nonalignment.py"]

doc_url = "https://pysan-dev.github.io/pysan/"

git_url = "https://github.com/pysan-dev/pysan/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: 20_multi_overview.ipynb (unless otherwise specified).

__all__ = ['plot_sequences']

# Cell
import matplotlib.pyplot as plt
import numpy as np
from ..core.elements import get_alphabet
def plot_sequences(sequences, gap=True):
    "Creates a scatter style sequence plot for a collection of sequences."
    max_sequence_length = max([len(s) for s in sequences])
    plt.figure(figsize=[max_sequence_length*0.3,0.3 * len(sequences)])

    for y,sequence in enumerate(sequences):
        np_sequence = np.array(sequence)
        alphabet_len = len(get_alphabet(sequence))

        plt.gca().set_prop_cycle(None)
        unique_values = get_alphabet(sequence)
        for i, value in enumerate(unique_values):

            if gap:
                points = np.where(np_sequence == value, y + 1, np.nan)
                plt.scatter(x=range(len(np_sequence)), y=points, marker='s', label=value, s=100)
            else:
                points = np.where(np_sequence == value, 1, np.nan)
                plt.bar(range(len(points)), points, bottom=[y for x in range(len(points))], width=1, align='edge', label=i)

    if gap:
        plt.ylim(0.4, len(sequences) + 0.6)
        plt.xlim(-0.6, max_sequence_length - 0.4)
    else:
        plt.ylim(0,len(sequences))
        plt.xlim(0,max_sequence_length)

    handles, labels = plt.gca().get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    plt.legend(by_label.values(), by_label.keys(), bbox_to_anchor=(1.0, 1.1), loc='upper left')
    plt.tick_params(
        axis='y',
        which='both',
        left=False,
        labelleft=False)

    return plt


# AUTOGENERATED! DO NOT EDIT! File to edit: 11_core_elements.ipynb (unless otherwise specified).

__all__ = ['get_alphabet', 'get_element_counts', 'get_first_positions', 'get_element_frequency', 'plot_element_counts']

# Cell
def get_alphabet(sequence):
    "Computes the alphabet of a given sequence (set of its unique elements)."
    return set(sequence)

# Cell
def get_element_counts(sequence):
    "Counts the number of occurances for each element in a sequence, returning a dictionary containing the elements as keys and their counts as values."
    alphabet = get_alphabet(sequence)

    counts = {}
    for element in alphabet:
        counts[element] = sequence.count(element)

    return counts

# Cell
def get_first_positions(sequence):
    "Reports the first occurance of each element in the sequence in a dictionary, with each element as keys, and their first position as values."
    unique_elements = list(set(sequence))

    first_positions = {}
    for element in unique_elements:
        first_positions[element] = sequence.index(element)

    return first_positions

# Cell
def get_element_frequency(sequence):
    "Computes the relative frequency (aka prevalence or unconditional probability) of each element in a sequence, returning a dictionary where each key is an element and each value is that elements relative frequency."
    alphabet = get_alphabet(sequence)

    prevalences = {}
    for element in alphabet:
        prevalences[element] = sequence.count(element) / len(sequence)

    return prevalences

# Cell
import matplotlib.pyplot as plt
def plot_element_counts(sequence):
    "Plots the number of occurances of each unique element in a given sequence."

    prev = get_element_counts(sequence)
    prev = {k: prev[k] for k in sorted(prev, key=prev.get)}

    xdata = [str(key) for key,value in prev.items()]
    ydata = [value for key,value in prev.items()]

    plt.figure(figsize=(len(xdata) * 2, len(ydata) * 0.5))
    plt.barh(xdata, ydata)
    plt.ylabel('Element')
    plt.xlabel('Number of Occurances')
    plt.gca().yaxis.grid(False)
    return plt


# AUTOGENERATED! DO NOT EDIT! File to edit: 13_core_ngrams.ipynb (unless otherwise specified).

__all__ = ['get_unique_ngrams', 'get_all_ngrams', 'get_ngram_universe', 'get_ngram_counts', 'plot_ngram_counts']

# Cell
def get_unique_ngrams(sequence, n):
    "Creates a list of all unique ngrams found in a given sequence."
    unique_ngrams = []
    for x in range(len(sequence) -  n + 1):
        this_ngram = sequence[x:x + n]

        if str(this_ngram) not in unique_ngrams:
            unique_ngrams.append(str(this_ngram))

    return [eval(x) for x in unique_ngrams]

# Cell
def get_all_ngrams(sequence, n):
    "Creates a list of all ngrams found in a given sequence."

    all_ngrams = []
    for x in range(len(sequence) -  n + 1):
        this_ngram = sequence[x:x + n]
        all_ngrams.append(this_ngram)

    return all_ngrams

# Cell
def get_ngram_universe(sequence, n):
    "Computes the universe of possible ngrams given a sequence. Where n is equal to the length of the sequence, the resulting number represents the sequence universe."
    # if recurrance is possible, the universe is given by k^t (SSA pg 68)
    k = len(set(sequence))
    if k > 10 and n > 10:
        return 'really big'
    return k**n

# Cell
def get_ngram_counts(sequence, n):
    "Computes the prevalence of ngrams in a sequence, returning a dictionary where each key is an ngram, and each value is the number of times that ngram appears in the sequence."

    ngrams = get_unique_ngrams(sequence, n)

    ngram_counts = {str(i):0 for i in ngrams}

    for x in range(len(sequence) -  n + 1):
        this_ngram = sequence[x:x + n]
        ngram_counts[str(this_ngram)] += 1

    return ngram_counts

# Cell
import matplotlib.pyplot as plt
def plot_ngram_counts(sequence, n):
    "Plots the number of occurances of ngrams in a given sequence."

    ngram_counts = get_ngram_counts(sequence, n)
    ngram_counts = {k: ngram_counts[k] for k in sorted(ngram_counts, key=ngram_counts.get)}

    xdata = [key[1:len(key)-1].replace(', ', ', ') for key,value in ngram_counts.items()]
    ydata = [value for key,value in ngram_counts.items()]

    plt.figure()
    plt.barh(xdata, ydata, label=str(n) +'-gram')
    plt.xlabel('Number of Occurances')
    plt.gca().yaxis.grid(False)
    plt.legend()
    return plt


# AUTOGENERATED! DO NOT EDIT! File to edit: 21_multi_attributes.ipynb (unless otherwise specified).

__all__ = ['are_recurrent', 'get_summary_statistic', 'get_routine_scores', 'get_synchrony', 'get_sequence_frequencies']

# Cell
def are_recurrent(sequences):
    "Returns true if any of the sequences in a given collection are recurrant, false otherwise."

    for sequence in sequences:
        if pysan_core.is_recurrent(sequence):
            return True

    return False

# Cell
def get_summary_statistic(sequence, function):
    "Computes a summary statistic (e.g. entropy, complexity, or turbulence) for each sequence in a collection, returning the results as a list."

    pass

# Cell
def get_routine_scores(sequences, duration):
    "Returns a list containing the routine scores for each sequence in a collection using :meth:`get_routine() <pysan.core.get_routine>`."

    pass

# Cell
def get_synchrony(sequences):
    "Computes the normalised synchrony between a two or more sequences. Synchrony here refers to positions with identical elements, e.g. two identical sequences have a synchrony of 1, two completely different sequences have a synchrony of 0. The value is normalised by dividing by the number of positions compared. This computation is defined in Cornwell's 2015 book on social sequence analysis, page 230."

    shortest_sequence = min([len(s) for s in sequences])

    same_elements = []
    for position in range(shortest_sequence):

        elements_at_this_position = []
        for sequence in sequences:
            elements_at_this_position.append(sequence[position])

        same_elements.append(elements_at_this_position.count(elements_at_this_position[0]) == len(elements_at_this_position))

    return same_elements.count(True) / shortest_sequence

# Cell
def get_sequence_frequencies(sequences):
    "Computes the frequencies of different sequences in a collection, returning a dictionary of their string representations and counts."

    # converting to strings makes comparison easy
    sequences_as_strings = [str(s) for s in sequences]

    sequence_frequencies = {}
    for sequence in set(sequences_as_strings):
        sequence_frequencies[sequence] = sequences_as_strings.count(sequence)

    sequence_frequencies = {k: v for k, v in sorted(sequence_frequencies.items(), key=lambda item: item[1], reverse=True)}

    return sequence_frequencies


# AUTOGENERATED! DO NOT EDIT! File to edit: 23_multi_edit_distances.ipynb (unless otherwise specified).

__all__ = ['get_optimal_distance', 'get_levenshtein_distance', 'get_hamming_distance']

# Cell
def get_optimal_distance(s1,s2, match = 0, mismatch = -1, gap = -1):
    "Computes the optimal matching distance between two sequences using the [Needleman-Wunsch algorithm](https://www.sciencedirect.com/science/article/abs/pii/0022283670900574?via%3Dihub) based on Devon Ryan's implementation found [here](<https://www.biostars.org/p/231391/)."

    penalty = {'MATCH': match, 'MISMATCH': mismatch, 'GAP': gap} #A dictionary for all the penalty valuse.
    n = len(s1) + 1 #The dimension of the matrix columns.
    m = len(s2) + 1 #The dimension of the matrix rows.
    al_mat = np.zeros((m,n),dtype = float) #Initializes the alighment matrix with zeros.
    p_mat = np.zeros((m,n),dtype = str) #Initializes the pointer matrix with zeros.
    #Scans all the first rows element in the matrix and fill it with "gap penalty"
    for i in range(m):
        al_mat[i][0] = penalty['GAP'] * i
        p_mat[i][0] = 'V'
    #Scans all the first columns element in the matrix and fill it with "gap penalty"
    for j in range (n):
        al_mat[0][j] = penalty['GAP'] * j
        p_mat [0][j] = 'H'


    #-------------------------------------------------------
    #This function returns to values for cae of match or mismatch
    def Diagonal(n1,n2,pt):
        if(n1 == n2):
            return pt['MATCH']
        else:
            return pt['MISMATCH']

    #------------------------------------------------------------
    #This function gets the optional elements of the aligment matrix and returns the elements for the pointers matrix.
    def Pointers(di,ho,ve):
        pointer = max(di,ho,ve) #based on python default maximum(return the first element).

        if(di == pointer):
            return 'D'
        elif(ho == pointer):
            return 'H'
        else:
             return 'V'

    #Fill the matrix with the correct values.
    p_mat [0][0] = 0 #Return the first element of the pointer matrix back to 0.
    for i in range(1,m):
        for j in range(1,n):
            di = al_mat[i-1][j-1] + Diagonal(s1[j-1],s2[i-1],penalty) #The value for match/mismatch -  diagonal.
            ho = al_mat[i][j-1] + penalty['GAP'] #The value for gap - horizontal.(from the left cell)
            ve = al_mat[i-1][j] + penalty['GAP'] #The value for gap - vertical.(from the upper cell)
            al_mat[i][j] = max(di,ho,ve) #Fill the matrix with the maximal value.(based on the python default maximum)
            p_mat[i][j] = Pointers(di,ho,ve)

    #print(np.matrix(al_mat))
    #print(np.matrix(p_mat))

    # optimal alignment score = bottom right value in al_mat
    score = al_mat[m-1][n-1]
    #print(score)
    if score == 0: # fixes -0 bug for completeness
        return 0

    return -score

# Cell
def get_levenshtein_distance(s1,s2):
    "Computes the [Levenshtein II distance](https://journals.sagepub.com/doi/abs/10.1177/0049124110362526) between two sequences, which is the optimal distance using only insertions and deletions. This is identical to the `get_optimal_distance` method with a mismatch cost of ~infinity (-9999999) and a gap cost of -1. See the `get_optimal_distance` method with its default parameters for the Levenshtein I distance."
    return get_optimal_distance(s1,s2, match=0, mismatch=-9999999, gap=-1)

# Cell
def get_hamming_distance(s1,s2):
    "Computes the Hamming distance  between two sequences, which is the optimal distance using only substitutions (no indels). This is identical to the `get_optimal_distance` method with a mismatch cost of -1 and a gap cost of ~infinity (-999999). Note that this can only be used on sequences of the same length given the infinite cost of gaps."
    if len(s1) != len(s2):
        raise Exception('sequences provided are not equal length - cannot compute Hamming distance')

    return get_optimal_distance(s1,s2, match=0, mismatch=-1, gap=-999999)


# AUTOGENERATED! DO NOT EDIT! File to edit: 12_core_subsequences.ipynb (unless otherwise specified).

__all__ = ['get_subsequences', 'get_ndistinct_subsequences']

# Cell
from itertools import combinations
def get_subsequences(sequence):
    subsequences = []
    for i in range(0, len(sequence)+1):
        temp = [list(x) for x in combinations(sequence, i)]
        if len(temp)>0:
            subsequences.extend(temp)
    return subsequences[1:]

# Cell
def get_ndistinct_subsequences(sequence):
    # this implementation works on strings, so parse non-strings to strings
    if sequence is not str:
        sequence = [str(e) for e in sequence]

    # create an array to store index of last
    last = [-1 for i in range(256 + 1)] # hard-coded value needs explaining -ojs

    # length of input string
    sequence_length = len(sequence)

    # dp[i] is going to store count of discount subsequence of length of i
    dp = [-2 for i in range(sequence_length + 1)]

    # empty substring has only one subseqence
    dp[0] = 1

    # Traverse through all lengths from 1 to n
    for i in range(1, sequence_length + 1):

        # number of subseqence with substring str[0...i-1]
        dp[i] = 2 * dp[i - 1]

        # if current character has appeared before, then remove all subseqences ending with previous occurrence.
        if last[ord(sequence[i - 1])] != -1:
            dp[i] = dp[i] - dp[last[ord(sequence[i - 1])]]

        last[ord(sequence[i - 1])] = i - 1

    return dp[sequence_length]


# AUTOGENERATED! DO NOT EDIT! File to edit: 10_core_overview.ipynb (unless otherwise specified).

__all__ = ['plot_sequence', 'plot_sequence_1d']

# Cell
from .elements import get_alphabet
import numpy as np, matplotlib.pyplot as plt

def plot_sequence(sequence, highlighted_ngrams=[]):
    "Creates a standard sequence plot where each element corresponds to a position on the y-axis. The optional highlighted_ngrams parameter can be one or more n-grams to be outlined in a red box."
    np_sequence = np.array(sequence)
    alphabet_len = len(get_alphabet(sequence))

    plt.figure(figsize=[len(sequence)*0.3,alphabet_len * 0.3])

    unique_values = list(set(sequence))
    for i, value in enumerate(unique_values):

        points = np.where(np_sequence == value, i, np.nan)

        plt.scatter(x=range(len(np_sequence)), y=points, marker='s', label=value, s=35)

    plt.yticks(range(len(unique_values)), unique_values)
    plt.ylim(-1, len(unique_values))

    # highlight any of the n-grams given

    if highlighted_ngrams != []:

        def highlight_ngram(ngram):
            n = len(ngram)
            match_positions = []
            for x in range(len(sequence) -  n + 1):
                this_ngram = sequence[x:x + n]
                if str(this_ngram) == str(ngram):
                    match_positions.append(x)

            for position in match_positions:
                bot = min(ngram) - 1.5
                top = max(ngram) - 0.5
                left = position - 0.5
                right = left + n

                line_width = 1
                plt.plot([left,right], [bot,bot], color='red', linewidth=line_width)
                plt.plot([left,right], [top,top], color='red', linewidth=line_width)
                plt.plot([left,left], [bot,top], color='red', linewidth=line_width)
                plt.plot([right,right], [bot,top], color='red', linewidth=line_width)

        # check if only one n-gram has been supplied
        if type(highlighted_ngrams[0]) is int:

            highlight_ngram(highlighted_ngrams)

        else: # multiple n-gram's found

            for ngram in highlighted_ngrams:
                highlight_ngram(ngram)

    return plt

# Cell
def plot_sequence_1d(sequence, flat=False):
    "Plots a sequence in one dimension - useful for stacking multiple sequences above one another."

    np_sequence = np.array(sequence)
    alphabet_len = len(get_alphabet(sequence))

    plt.figure(figsize=[len(sequence)*0.4, 0.5])

    unique_values = list(set(sequence))
    for i, value in enumerate(unique_values):

        points = np.where(np_sequence == value, 1, np.nan)

        plt.bar(range(len(points)), points, width=1, align='edge', label=i)

    plt.ylim(-0.3, 1.3)
    plt.tick_params(
        axis='y',
        which='both',
        left=False,
        labelleft=False)
    plt.xlabel('Position, p')
    plt.legend(bbox_to_anchor=(1, 1.2), loc='upper left')

    return plt


# AUTOGENERATED! DO NOT EDIT! File to edit: 22_multi_derivatives.ipynb (unless otherwise specified).

__all__ = ['get_motif', 'get_modal_state']

# Cell
def get_motif(sequences):
    "Computes the motif for a given collection of sequences. A motif is a representative sequence for all sequences in a collection, with blank values (0) being those which are variable within the collection, and fixed values which are not. Motifs are related to the measure of synchrony (see `get_synchrony`) in that synchrony is equal to the number of non-blank elements in the motif."

    shortest_sequence = min([len(s) for s in sequences])

    same_elements = []
    for position in range(shortest_sequence):

        elements_at_this_position = []
        for sequence in sequences:
            elements_at_this_position.append(sequence[position])

        if elements_at_this_position.count(elements_at_this_position[0]) == len(elements_at_this_position):
            same_elements.append(sequences[0][position])
        else:
            same_elements.append(0)

    return same_elements

# Cell
def get_modal_state(sequences):
    "Computes the modal states for each position in a collection of sequences, returning a sequence of tuples containing the modal element and its number of occurances at that position."

    longest_sequence = max([len(s) for s in sequences])

    modal_elements = []
    for position in range(longest_sequence):

        elements_at_this_position = []
        for sequence in sequences:
            try:
                elements_at_this_position.append(sequence[position])
            except:
                continue

        # this line leaves multi-modal position behaviour undefined
        modal_element = max(set(elements_at_this_position), key=elements_at_this_position.count)

        modal_elements.append((modal_element, elements_at_this_position.count(modal_element)))

    return modal_elements


# AUTOGENERATED! DO NOT EDIT! File to edit: 14_core_transitions.ipynb (unless otherwise specified).

__all__ = ['get_transitions', 'get_ntransitions', 'get_transition_matrix', 'plot_transition_matrix']

# Cell
def get_transitions(sequence):
    "Extracts a list of transitions from a sequence, returning a list of lists containing each transition."

    transitions = []
    for position in range(len(sequence) - 1):
        if sequence[position] != sequence[position + 1]:
            transitions.append([sequence[position], sequence[position + 1]])

    return transitions

# Cell
def get_ntransitions(sequence):
    "Computes the number of transitions in a sequence."	
    return len(get_transitions(sequence))

# Cell
from .elements import get_alphabet
from .ngrams import get_all_ngrams
import numpy as np, pandas as pd
def get_transition_matrix(sequence, alphabet=None, verbose=False):
    "Computes a transition matrix for each bigram in a sequence. The resulting matrix can be interpreted by reading along the side first, then across the top, indicating from the element down the side, to the element along the top. For example, to find the number of transitions from element 2 to element 3, find element 2 down the side, then follow that row across until it reaches element 3 across the top."
    if alphabet == None:
        alphabet = get_alphabet(sequence)
    all_ngrams = get_all_ngrams(sequence, 2)

    transition_matrix = np.zeros((len(alphabet), len(alphabet)))
    descriptive_matrix = [['-' for x in range(len(alphabet))] for y in range(len(alphabet))]

    for x, element_row in enumerate(alphabet):
        for y, element_column in enumerate(alphabet):
            current_ngram = [element_row, element_column]
            descriptive_matrix[x][y] = 'n' + str(current_ngram)
            #print('from', current_ngram[0], 'to', current_ngram[1], ':', all_ngrams.count(current_ngram))
            transition_matrix[x, y] = all_ngrams.count(current_ngram)

    # add column & index labelling in TraMineR style
    pre_alphabet = [str(a) + '->' for a in alphabet]
    post_alphabet = ['->' + str(a) for a in alphabet]

    if verbose:
        de_df = pd.DataFrame(descriptive_matrix, columns=post_alphabet, index=pre_alphabet)
        print(de_df)
    tm_df = pd.DataFrame(transition_matrix, columns=post_alphabet, index=pre_alphabet)
    return tm_df

# Cell
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import copy

def plot_transition_matrix(sequence, cmap='summer'):
    "Computes and plots a transition matrix, returning a colored matrix with elements at position n up the y axis, and elements at position n+1 along the x axis."

    matrix = get_transition_matrix(sequence)

    results_size = len(matrix.columns)
    values = np.empty((results_size, results_size), dtype=object)

    for r, row in enumerate(matrix.values):
        for e, element in enumerate(row):
            if element == "-":
                values[r, e] = 100
                continue
            if element == "":
                values[r, e] = np.nan
                continue
            if "*" in str(element):
                value = element.replace("*", "")
                values[r, e] = float(value)
            else:
                values[r, e] = element

    current_cmap = copy.copy(cm.get_cmap(cmap))
    current_cmap.set_bad(color="white")

    plt.figure()

    # this one-lines sets the x axis to appear at the top of this plot only
    with plt.rc_context({'xtick.bottom':False, 'xtick.labelbottom':False, 'xtick.top':True, 'xtick.labeltop':True}):
        ax = plt.gca()
        ax.xaxis.set_label_position('top')
        plt.imshow(np.array(values).astype(np.float), cmap=current_cmap)

        plt.yticks(range(len(matrix.index)), list(matrix.index))
        plt.xticks(range(len(matrix.columns)), list(matrix.columns))
        cbar = plt.colorbar()
        #cbar.set_ticks([-100, -80, -60, -40, -20, 0, 20, 40, 60, 80, 100])
        #cbar.set_ticklabels([-1, -0.8, -0.6, -0.4, -0.2, 0, 0.2, 0.4, 0.6, 0.8, 1])
        plt.grid(False)
        return plt


# AUTOGENERATED! DO NOT EDIT! File to edit: 24_multi_nonalignment.ipynb (unless otherwise specified).

__all__ = ['get_combinatorial_distance']

# Cell
def get_combinatorial_distance(s1,s2):
    "Computes the combinatorial distance between two sequences. This is defined as 1 minus the number of common subsequences divided by the square root of the product of the number of subsequences in each sequence. See page 149 in Social Sequence Analysis by Benjamin Cornwell for details."

    s1_subs = pysan_core.get_subsequences(s1)
    s2_subs = pysan_core.get_subsequences(s2)

    # parse to strings so that they can be easily compared
    # - haven't tried it without, may be faster...
    s1_subs_strings = [str(s) for s in s1_subs]
    s2_subs_strings = [str(s) for s in s2_subs]

    common_subs = list(set(s1_subs_strings) & set(s2_subs_strings))

    bottom_fraction = math.sqrt(len(s1_subs) * len(s2_subs))
    full_fraction = len(common_subs) / bottom_fraction

    return 1 - full_fraction


# AUTOGENERATED! DO NOT EDIT! File to edit: 15_core_spells.ipynb (unless otherwise specified).

__all__ = ['get_spells', 'get_longest_spell', 'get_spell_durations']

# Cell
def get_spells(sequence):
    "Returns a list of tuples where each tuple holds the element and the length of the spell (also known as run or episode) for each spell in the sequence."

    # get each spell and its length
    spells = [(k, sum(1 for x in v)) for k,v in itertools.groupby(sequence)]
    # this is functionally equivalent to the following;
    # spells = [(k, len(list(v))) for k,v in itertools.groupby(sequence)]

    return spells

# Cell
def get_longest_spell(sequence):
    "Returns a dict containing the element, count, and starting position of the longest spell in the sequence. The keys of this dict are 'element, 'count', and 'start'."

    spells = get_spells(sequence)

    longest_spell = max(count for element, count in spells)

    for i, (element, count) in enumerate(spells):
        if count == longest_spell:
            # sum the counts of all previous spells to get its starting position
            position_in_sequence = sum(count for _,count in spells[:i])

            return {'element':element, 'count':count,'start':position_in_sequence}

# Cell
def get_spell_durations(sequence):
    "Computes the durations of each spell in the sequence, returning a list."

    spells = get_spells(sequence)
    durations = [spell[1] for spell in spells]

    return durations


# AUTOGENERATED! DO NOT EDIT! File to edit: 16_core_statistics.ipynb (unless otherwise specified).

__all__ = ['is_recurrent', 'get_entropy', 'get_turbulence', 'get_complexity', 'get_routine']

# Cell
def is_recurrent(sequence):
    "Returns true if the given sequence is recurrent (elements can exist more than once), otherwise returns false."	
    element_counts = get_element_counts(sequence)

    truths = [count > 1 for element, count in element_counts.items()]

    if True in truths:
        return True
    return False

# Cell
def get_entropy(sequence):
    "Computes the normalised [Shannon entropy](https://en.wikipedia.org/wiki/Entropy_(information_theory)) of a given sequence, using the [scipy.stats.entropy](https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.entropy.html) implementation. Note that this measure is insensitive to transition frequency or event order, so should be used in conjunction with other measures."

    alphabet = get_alphabet(sequence)

    entropy = 0
    for state in alphabet:
        proportion_occurances = sequence.count(state) / len(sequence)
        entropy += proportion_occurances * math.log(proportion_occurances)

    maximal_occurances = 1 / len(alphabet)
    alphabet_entropy = sum([maximal_occurances * math.log(maximal_occurances) for x in alphabet])

    if alphabet_entropy == 0:
        return 0

    return -entropy / -alphabet_entropy

# Cell
def get_turbulence(sequence):
    "Computes turbulence for a given sequence, based on [Elzinga & Liefbroer's 2007 definition](https://www.researchgate.net/publication/225402919_De-standardization_of_Family-Life_Trajectories_of_Young_Adults_A_Cross-National_Comparison_Using_Sequence_Analysis) which is also implemented in the [TraMineR](http://traminer.unige.ch/doc/seqST.html) sequence analysis library."

    phi = get_ndistinct_subsequences(sequence)

    #print('phi', phi)

    state_durations = [value for key, value in get_spells(sequence)]

    #print('durations', state_durations)
    #print('mean duration', statistics.mean(state_durations))

    variance_of_state_durations = statistics.variance(state_durations)

    #print('variance', variance_of_state_durations)

    tbar = statistics.mean(state_durations)

    maximum_state_duration_variance = (len(sequence) - 1) * (1 - tbar) ** 2

    #print('smax', maximum_state_duration_variance)

    top_right = maximum_state_duration_variance + 1
    bot_right = variance_of_state_durations + 1

    turbulence = math.log2(phi * (top_right / bot_right))

    #print('turbulence', turbulence)

    return turbulence

# Cell
def get_complexity(sequence):
    "Computes the complexity of a given sequence, based on TraMineR's [seqici](http://traminer.unige.ch/doc/seqici.html) method."

    alphabet = get_alphabet(sequence)

    pre_log = 1 / len(alphabet)
    hmax = -math.log(pre_log)
    #print('hmax', hmax)
    if hmax == 0:
        return 0 # all identical elements, no complexity

    hs = get_entropy(sequence)
    #print('hs', hs)

    qs = get_ntransitions(sequence)
    #print('qs', qs)

    qmax = len(sequence) - 1
    #print('qmax', qmax)

    norm_transitions = qs / qmax
    norm_entropy = hs / hmax

    #print('nt', norm_transitions)
    #print('ne', norm_entropy)

    complexity = math.sqrt(norm_transitions * norm_entropy)

    #print('complexity', complexity)
    return complexity

# Cell
def get_routine(sequence, duration):
    "Computes a normalised measure of routine within a sequence for a given duration within that sequence. E.g. with a sequence where each element is one day, calling get_routine() with a duration of 7 would look at weekly routines. Note that this routine measure is identical to the multisequence measure of synchrony, but applied within-sequence in duration length chunks."

    if len(sequence) % duration != 0:
        raise Exception('sequence not divisible by interval, check data input')

    num_cycles = int(len(sequence) / duration)
    cycles = [sequence[n * duration:n * duration + duration] for n in range(num_cycles)]

    return pysan_ms.get_synchrony(cycles)